#!/usr/bin/env python

"""
A really simple virtual processor employing a simple set of instructions which
ignore low-level operations and merely concentrate on variable access, structure
access, structure allocation and function invocations.

Copyright (C) 2007, 2008 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The execution model of the virtual processor involves the following things:

  * Memory
  * PC (program counter) stack
  * Frame stack (containing invocation frames in use and in preparation plus
                temporary storage)
  * Local frame pointer stack
  * Invocation frame pointer stack
  * Exception handler stack
  * Registers: current value, boolean status value, source value, result,
               current exception

The memory contains constants, global variable references and program code.

The PC stack contains the return address associated with each function
invocation.

The frame pointer stack tracks the position of frames within the frame stack.
"""

class IllegalInstruction(Exception):
    pass

class IllegalAddress(Exception):
    pass

class EmptyPCStack(Exception):
    pass

class EmptyFrameStack(Exception):
    pass

class RSVPMachine:

    "A really simple virtual processor."

    def __init__(self, memory, objtable, paramtable, pc=None, debug=0):

        """
        Initialise the processor with a 'memory' (a list of values containing
        instructions and data) and the optional program counter 'pc'.
        """

        self.memory = memory
        self.objtable = objtable
        self.paramtable = paramtable
        self.pc = pc or 0
        self.debug = debug

        # Stacks.

        self.pc_stack = []
        self.frame_stack = []
        self.local_sp_stack = []
        self.invocation_sp_stack = []
        self.handler_stack = []

        # Registers.

        self.instruction = None
        self.operand = None
        self.value = None
        self.status = None
        self.source = None
        self.result = None
        self.exception = None

    def load(self, address):

        "Return the value at the given 'address'."

        try:
            return self.memory[address]
        except IndexError:
            raise IllegalAddress, address

    def save(self, address, value):

        "Save to the given 'address' the specified 'value'."

        try:
            self.memory[address] = value
        except IndexError:
            raise IllegalAddress, address

    def new(self, size):

        """
        Allocate space of the given 'size', returning the address of the space.
        """

        addr = len(self.memory)
        for i in range(0, size):
            self.memory.append(None)
        return addr

    def push_pc(self, data):

        "Push 'data' onto the PC stack."

        self.pc_stack.append(data)

    def pull_pc(self):

        "Pull a value from the PC stack and return it."

        try:
            return self.pc_stack.pop()
        except IndexError:
            raise EmptyPCStack

    def run(self):

        "Execute code in the memory, starting from the current PC address."

        try:
            while 1:
                self.execute()
        except EmptyPCStack:
            pass

    def execute(self):

        "Execute code in the memory at the current PC address."

        self.instruction = self.load(self.pc)
        self.operand = self.instruction.get_operand()

        instruction_name = self.instruction.__class__.__name__
        if self.debug:
            print "%8d %s" % (self.pc, instruction_name)

        method = self.get_method(instruction_name)

        # Process any inputs of the instruction.

        self.process_inputs()
        next_pc = method()

        # Update the program counter.

        if next_pc is None:
            self.pc += 1
        else:
            self.pc = next_pc

    def get_method(self, instruction_name):

        "Return the handler method for the given 'instruction_name'."

        method = getattr(self, instruction_name, None)
        if method is None:
            raise IllegalInstruction, (self.pc, instruction_name)
        return method

    def process_inputs(self):

        """
        Process any inputs of the current instruction. This permits any directly
        connected sub-instructions to produce the effects that separate
        instructions would otherwise have.
        """

        for input in (self.instruction.input, self.instruction.source):
            if input is not None:
                method = self.get_method(input)
                method()

    def jump(self, addr, next):

        """
        Jump to the subroutine at (or identified by) 'addr'. If 'addr'
        identifies a library function then invoke the library function and set
        PC to 'next' afterwards; otherwise, set PC to 'addr'.
        """

        if isinstance(addr, str):
            getattr(self, addr)()
            return next
        else:
            self.push_pc(self.pc + 2)
            return addr

    # Instructions.

    def LoadConst(self):
        self.value = None, self.operand

    def LoadName(self):
        frame = self.local_sp_stack[-1]
        self.value = self.frame_stack[frame + self.operand]

    def StoreName(self):
        frame = self.local_sp_stack[-1]
        self.frame_stack[frame + self.operand] = self.value

    LoadTemp = LoadName
    StoreTemp = StoreName

    def LoadAddress(self):
        # Preserve context (potentially null).
        self.value = self.load(self.operand)

    def LoadAddressContext(self):
        value = self.load(self.operand)
        # Replace the context with the current value.
        self.value = self.value[1], value[1]

    def StoreAddress(self):
        # Preserve context.
        self.save(self.operand, self.value)

    def MakeObject(self):
        # Introduce null context for new object.
        self.value = None, self.new(self.operand)

    def LoadAttr(self):
        context, ref = self.value
        loaded_context, loaded_ref = self.load(ref + self.operand)
        if loaded_context is None:
            # Override null context with owning instance.
            self.value = ref, loaded_ref
        else:
            self.value = loaded_context, loaded_ref

    def StoreAttr(self):
        context, ref = self.value
        self.save(ref + self.operand, self.source)

    def LoadAttrIndex(self):
        context, ref = self.value
        code = self.load(ref) # + 0 (the classcode)
        element = self.objtable[code + self.operand]
        found_code, found_attr = element
        if found_code == code:
            # NOTE: The found context should be tested against the object's context.
            # NOTE: Compatibility between contexts should be stored in the table, or
            # NOTE: an incompatible context should be preserved.
            self.value = ref, found_ref
        else:
            # NOTE: This should cause an attribute error.
            self.value = None

    def StoreAttrIndex(self): pass

    def MakeFrame(self):
        self.invocation_sp_stack.append(len(self.frame_stack))
        self.frame_stack.extend([None] * self.operand)

    def JumpWithFrame(self):
        self.local_sp_stack.append(self.invocation_sp_stack[-1]) # adopt the invocation frame
        context, ref = self.value
        return self.jump(ref.code_location, self.pc + 1) # return to the instruction after this one

    def DropFrame(self):
        result = self.pull()
        self.local_sp_stack.pop()
        frame = self.invocation_sp_stack.pop()
        self.frame_stack = self.frame_stack[:frame] # reset stack before call

    def CheckFrame(self): pass

    def LoadCallable(self): pass

    def LoadContext(self):
        context, ref = self.value
        self.push((None, context))

    def CheckSelf(self): pass

    def CheckContext(self): pass

    def Return(self):
        self.pc = self.pull_pc()

    def LoadResult(self):
        self.value = self.result

    def StoreResult(self):
        self.result = self.value

    def LoadException(self):
        self.value = self.exception

    def StoreException(self):
        self.exception = self.value

    def LoadBoolean(self):
        self.value = self.status

    def Jump(self):
        return self.operand

    def JumpIfTrue(self):
        if self.status:
            return self.operand

    def JumpIfFalse(self):
        if not self.status:
            return self.operand

    def StoreFrame(self):
        frame = self.invocation_sp_stack[-1] # different from the current frame after MakeFrame
        self.frame_stack[frame + self.operand] = self.value

# vim: tabstop=4 expandtab shiftwidth=4
